/*
   Copyright (c) 2004-2005 The University of Chicago, as Operator
   of Argonne National Laboratory.
*/
package gov.anl.aps.irmis.apps.ioc.cfw;

import java.net.URL;
import java.util.List;
import java.util.Set;
import java.util.Iterator;
import javax.swing.SwingUtilities;
import javax.swing.JDesktopPane;

// XAL framework stuff
import gov.sns.application.*;

// work queue support from concurrent.jar
import EDU.oswego.cs.dl.util.concurrent.QueuedExecutor;

// general app utilities
import gov.anl.aps.irmis.apps.SwingThreadWork;

// rdbCore imports
import gov.anl.aps.irmis.persistence.pv.IOCDAO;
import gov.anl.aps.irmis.persistence.pv.IOC;
import gov.anl.aps.irmis.persistence.DAOException;


/**
 * The primary controller class for IRMIS IOC application.
 * Provides "action" methods to to query the IRMIS database for ioc's and
 * related data. These action methods are typically invoked by the 
 * <code>IOCWindow</code>.
 * Any time-consuming activity should be queued up using the 
 * <code>queuedExecutor</code>, which is similar to the SwingWorker concept.
 */
public class IOCDocument extends XalInternalDocument {

	/** The main data model of this application */
	private IOCModel _model;

    /** 
     * Background work queue we use to handle lengthy logic (ie. sql queries). This
     * is a work queue backed by a single common background thread. It's important
     * to conduct most if not all sql activity on this thread, as the Hibernate
     * session is maintained in ThreadLocal storage, and therefore should only
     * exist on one thread to take advantage of Hibernate caching. The QueuedExecutor
     * class comes from the public domain concurrent.jar library, most of which
     * went into Sun's JDK 1.5 java.util.concurrent package. We're still on JDK1.4.
     */
	QueuedExecutor queuedExecutor = new QueuedExecutor();
	

	/** Create a new empty document */
    public IOCDocument() {
        this(null);
    }

	/** Create a new document from url */
    public IOCDocument(final URL url) {
        //setSource(url);
        _model = new IOCModel();

        // don't have anything here yet
    }

    public void saveDocumentAs(final URL url) {
        return;
    }

    /**
     * Specifies the document file suffixes supported by this document.
     * @return An array of file suffixes corresponding to writable files
     */
    public String[] writableDocumentTypes() {
        return new String[] {};
	}

    /**
	 * Overriding default window title to my liking.
     * @param newTitle The new title for this document.
     */
    public void setTitle( final String newTitle ) {
        String modifiedTitle = "IOC Viewer" + newTitle;
		super.setTitle( modifiedTitle );
        _documentListenerProxy.titleChanged( this, title );
    }	

    /**
     * Queue up job for background thread to gather the current set of IOC
     * objects. When complete, fire off <code>IOCModelEvent</code> to notify GUI
     * that new list of iocs is available for display. 
     */
    public void loadInitialIOCData() {
        try {
            queuedExecutor.execute(new SwingThreadWork(myWindow()) {

                    public void doNonUILogic() {
                        // Get current set of IOC objects, and put into IOCModel
                        IOCDAO iocDAO = null;
                        List iocList = null;
                        try {
                            iocDAO = new IOCDAO();
                            iocList = iocDAO.findIOCs();
                        } catch (DAOException de) {
                            de.printStackTrace();
                            // log error
                        }
                        _model.setIocList(iocList);
                    }

                    public void doUIUpdateLogic() {
                        // fire off notification of modification to listeners
                        _model.notifyIOCModelListeners(new IOCModelEvent(IOCModelEvent.NEW_IOC_LIST));
                    }
                });
        } catch (InterruptedException ie) {}
    }
            
    /**
     * Make a main window by instantiating the my custom window.
     */
    public void makeMainWindow() {
        _mainWindow = new IOCWindow(this);
        loadInitialIOCData();
    }

    /**
     * Programatically selects an ioc for display. Has the same effect as a user
     * clicking an ioc from the ioc list in the GUI. Provides means for external
     * document to command us to display a particular ioc.
     *
     * @param iocName ioc to select
     */
    public void actionSelectIoc(String iocName) {
        final String selectedIocName = iocName;
        try {
            queuedExecutor.execute(new SwingThreadWork(myWindow()) {

                    public void doNonUILogic() {
                        // find IOC object by given ioc name
                        List iocList = _model.getIocList();
                        IOC selectedIoc = null;
                        if (iocList != null) {
                            Iterator iocIt = iocList.iterator();
                            while (iocIt.hasNext()) {
                                IOC ioc = (IOC)iocIt.next();
                                if (ioc.getIocName().equals(selectedIocName)) {
                                    selectedIoc = ioc;
                                    break;
                                }
                            }
                        }
                        _model.setSelectedIoc(selectedIoc);
                    }
                    
                    public void doUIUpdateLogic() {
                        // fire off notification of modification to listeners
                        _model.notifyIOCModelListeners(new IOCModelEvent(IOCModelEvent.NEW_IOC_SELECTION));
                    }
                });
        } catch (InterruptedException ie) {}            
    }

    /**
     * Queue up job for background thread to gather the current set of IOC
     * attributes. When complete, fire off <code>IOCModelEvent</code> to notify GUI
     * that new set of ioc attributes is available for display. 
     */
    public void actionAttributeSearch() {
        try {
            queuedExecutor.execute(new SwingThreadWork(myWindow()) {

                    public void doNonUILogic() {
                        IOC selectedIoc = _model.getSelectedIoc();

                        // get attributes explicitly (lazy data)
                        if (selectedIoc != null) 
                            selectedIoc.getAttributes();
                    }
                    
                    public void doUIUpdateLogic() {
                        // fire off notification of modification to listeners
                        _model.notifyIOCModelListeners(new IOCModelEvent(IOCModelEvent.NEW_IOC_ATTR_LIST));
                    }
                });
        } catch (InterruptedException ie) {}
    }


	/**
	 * Get the launch model which represents the main model of this document
	 * @return the main model of this document
	 */
	public IOCModel getModel() {
		return _model;
	}

    private IOCWindow myWindow() {
        return (IOCWindow)_mainWindow;
    }
}
