/*
   Copyright (c) 2004-2005 The University of Chicago, as Operator
   of Argonne National Laboratory.
*/
package gov.anl.aps.irmis.apps.ioc.cfw;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.Set;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.border.TitledBorder;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;

import gov.sns.application.*;

// rdbCore imports
import gov.anl.aps.irmis.persistence.pv.IOC;
import gov.anl.aps.irmis.persistence.pv.IOCAttribute;

/**
 * Primary GUI for IRMIS IOC application. All window layout, and Swing event
 * listeners are done here. Business logic work is delegated to <code>IOCDocument</code>,
 * which in turn requests that the IOCModel notify us here of changes to the data. In short,
 * we listen for <code>IOCModelEvent</code> here.
 */
public class IOCWindow extends XalInternalWindow {

	/** The main model for the document */
	final protected IOCModel _model;

    final protected IOCDocument document;

    // ordered array of attribute results for selected IOC
    private IOCAttribute[] attrArray;

	// Swing GUI components
    private JPanel topPanel;
    private JPanel iocListPanel;
    private JPanel selectedIocPanel;
    private JPanel iocDetailsPanel;
    private JPanel iocAttrPanel;

    private JSplitPane lrSplitPane;

    private JTable iocList;
    private JTable attrList;

    private JButton attrEditButton;
    private JButton attrAddButton;
    private JButton attrRemoveButton;
	
	/** 
	 * Creates a new instance of ComponentWindow
	 * @param aDocument The document for this window
	 */
    public IOCWindow(final XalInternalDocument aDocument) {
        super(aDocument);
        document = (IOCDocument)aDocument;

		_model = document.getModel();

        // make IOCWindow a listener for changes in IOCModel
		_model.addIOCModelListener( new IOCModelListener() {
                public void modified(IOCModelEvent e) {
                    updateView(e);
                }
            });

        // initial application window size
        setSize(400, 205);

        // build contents
		makeContents();
    }

    /**
     * Top-level method to build up Swing GUI components.
     */
    public void makeContents() {

        // scrollbar policy
        int hsbp = JScrollPane.HORIZONTAL_SCROLLBAR_NEVER;
        int vsbp = JScrollPane.VERTICAL_SCROLLBAR_ALWAYS;

        // topmost panel
        topPanel = new JPanel();
        topPanel.setLayout(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        c.weightx = 1.0;
        c.weighty = 1.0;
        c.fill = GridBagConstraints.BOTH;
        c.gridx = 0;
        c.gridy = 0;

        // ioc list panel
        iocListPanel = new JPanel(new BorderLayout());
        TitledBorder iocListTitle = BorderFactory.createTitledBorder("IOCs");
        iocListPanel.setBorder(iocListTitle);

        // ioc list table
        iocList = new JTable(new IOCResultsTableModel());
        iocList.getColumnModel().getColumn(0).setCellRenderer(new IOCTableCellRenderer());
    	JScrollPane iocListScroller = new JScrollPane(iocList, vsbp, hsbp);
        iocListPanel.add(iocListScroller);
        //iocList.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        iocList.getColumnModel().getColumn(0).setPreferredWidth(120);
        iocList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        // handle row selection in ioc list table
        iocList.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    if (e.getValueIsAdjusting()) return;
                    ListSelectionModel lsm = (ListSelectionModel)e.getSource();
                    if (!lsm.isSelectionEmpty()) {
                        int selectedIOCResultsRow = lsm.getMinSelectionIndex();
                        List iocList;
                        iocList = _model.getIocList();
                        IOC selectedIoc = (IOC)iocList.get(selectedIOCResultsRow);
                        _model.setSelectedIoc(selectedIoc);
                        document.actionAttributeSearch();
                    }
                }
            });

        // make the panel used to display data after ioc is selected
        selectedIocPanel = makeSelectedIocPanelContents();

        // split between ioc list and ioc details info on right
        lrSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,iocListPanel,selectedIocPanel);
        lrSplitPane.setDividerLocation(120);

        // add split panes to topPanel
        topPanel.add(lrSplitPane,c);

    	getContentPane().add(topPanel);
    }

    private JPanel makeSelectedIocPanelContents() {
        
        // scrollbar policy
        int hsbp = JScrollPane.HORIZONTAL_SCROLLBAR_NEVER;
        int vsbp = JScrollPane.VERTICAL_SCROLLBAR_ALWAYS;

        // ioc details panel containing ioc attribute info
        iocDetailsPanel = new JPanel();
        iocDetailsPanel.setLayout(new BoxLayout(iocDetailsPanel,BoxLayout.PAGE_AXIS));

        // ioc attributes panel
        iocAttrPanel = new JPanel(new BorderLayout());
        TitledBorder iocAttrTitle = BorderFactory.createTitledBorder("IOC Attributes");
        iocAttrPanel.setBorder(iocAttrTitle);

        // ioc attributes table
        attrList = new JTable(new IOCAttrTableModel());
        attrList.getColumnModel().getColumn(1).setCellRenderer(new IOCAttrTableCellRenderer());
    	JScrollPane attrListScroller = new JScrollPane(attrList, vsbp, hsbp);
        iocAttrPanel.add(attrListScroller);
        attrList.getColumnModel().getColumn(1).setPreferredWidth(50);
        attrList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        // handle row selection in attribute list table
        attrList.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    if (e.getValueIsAdjusting()) return;
                    ListSelectionModel lsm = (ListSelectionModel)e.getSource();
                }
            });

        // attributes action button bar
        JPanel attrButtons = new JPanel(new FlowLayout(FlowLayout.LEFT));
        attrEditButton = new JButton("Edit");
        attrEditButton.setEnabled(false);
        attrAddButton = new JButton("Add");
        attrAddButton.setEnabled(false);
        attrRemoveButton = new JButton("Remove");
        attrRemoveButton.setEnabled(false);
        attrButtons.add(attrEditButton);
        attrButtons.add(attrAddButton);
        attrButtons.add(attrRemoveButton);
        

        // add attribute subpanel to ioc details panel
        iocDetailsPanel.add(iocAttrPanel);
        iocDetailsPanel.add(attrButtons);

        return iocDetailsPanel;
    }

    /**
     * Update graphical view of IOCModel data based on event type.
     *
     * @param event type of data model change
     */
    public void updateView(IOCModelEvent event) {

        switch(event.getType()) {
            
        case IOCModelEvent.NEW_IOC_LIST: {
            IOCResultsTableModel tableModel = (IOCResultsTableModel)iocList.getModel();
            // request that ioc results table update itself
            tableModel.fireTableDataChanged();
            if (_model.getIocList() != null && _model.getIocList().size() > 0) {
                iocList.setRowSelectionInterval(0,0);  // preselect row 0
            }
            iocListPanel.validate();
            break;
        }

        case IOCModelEvent.NEW_IOC_ATTR_LIST: {
            IOC selectedIoc = _model.getSelectedIoc();
            // convert attribute set to an ordered array
            Set attributes = selectedIoc.getAttributes();
            Iterator attrIt = attributes.iterator();
            List tempList = new ArrayList();
            while (attrIt.hasNext()) {
                IOCAttribute attr = (IOCAttribute)attrIt.next();
                tempList.add(attr);
            }
            Collections.sort(tempList);
            attrArray = (IOCAttribute[])tempList.toArray(new IOCAttribute[0]);
            IOCAttrTableModel iocAttrTableModel = 
                (IOCAttrTableModel)attrList.getModel();
            iocAttrTableModel.fireTableDataChanged();
            break;
        }

        case IOCModelEvent.NEW_IOC_SELECTION: {
            List iocs = _model.getIocList();
            IOC selectedIoc = _model.getSelectedIoc();
            int row = -1;
            if (iocs != null) {
                Iterator iocIt = iocs.iterator();
                while (iocIt.hasNext()) {
                    row++;
                    IOC ioc = (IOC)iocIt.next();
                    if (ioc.getIocName().equals(selectedIoc.getIocName())) {
                        break;
                    }
                }                
            }
            iocList.changeSelection(row,0,false,false);
            break;
        }

        default: {}
        }
            
    }

    /********************************************************************
     * Inner classes supporting Swing components
     ********************************************************************/

    /**
     * Provides methods to get ioc results table cell data from
     * actual <code>IOCModel</code>. 
     */
    class IOCResultsTableModel extends AbstractTableModel {
        private String[] columnNames = {"IOC Name"};
        public int getColumnCount() {
            return columnNames.length;
        }
        public int getRowCount() {
            int size = 0;
            if (_model.getIocList() == null)
                return size;
            size = _model.getIocList().size();
            return size;
        }
        public String getColumnName(int col) {
            return columnNames[col];
        }
        public Object getValueAt(int row, int col) {
            List iocList = null;

            iocList = _model.getIocList();

            if (iocList != null) {            
                IOC ioc = (IOC)iocList.get(row);
                return ioc.getIocName();
            } else {
                return null;
            }
        }
        public Class getColumnClass(int c) {
            return getValueAt(0,c).getClass();
        }
    }

    /**
     * Provides methods to get ioc attribute table cell data from actual
     * <code>IOCModel</code>. See Swing tutorial.
     */
    class IOCAttrTableModel extends AbstractTableModel {
        private String[] columnNames = {"Name","Value"};
        public int getColumnCount() {
            return columnNames.length;
        }
        public int getRowCount() {
            int size = 0;
            if (attrArray != null) {
                size = attrArray.length;
            }
            return size;
        }
        public String getColumnName(int col) {
            return columnNames[col];
        }
        public Object getValueAt(int row, int col) {

            if (attrArray != null) {
                IOCAttribute attr = attrArray[row];
                if (col == 0) {
                    return attr.getAttributeType().getAttributeType();
                } else {
                    return attr.getAttributeValue();
                }
            } else {
                return null;
            }

        }
        public Class getColumnClass(int c) {
            return getValueAt(0,c).getClass();
        }
    }

    /**
     * Customized table cell renderer used for ioc table. It colors
     * the cell background and text foreground based on whether ioc is active
     * or not.
     */
    class IOCTableCellRenderer extends DefaultTableCellRenderer {
        public Component getTableCellRendererComponent (JTable table,
                                                        Object value,
                                                        boolean selected,
                                                        boolean focused,
                                                        int row, int column) {
            setEnabled(table==null || table.isEnabled());
            if (column == 0) {
                List iocList = _model.getIocList();
                if (iocList != null) {
                    IOC ioc = (IOC)iocList.get(row);
                    if (!ioc.getActive()) {
                        setBackground(Color.lightGray);
                        setToolTipText("inactive");
                    } else {
                        setBackground(Color.white);
                    }
                    setForeground(Color.blue);
                }
            }
            super.getTableCellRendererComponent(table,value,selected,focused,row,column);
            return this;
        }
    }

    class IOCAttrTableCellRenderer 
        implements TableCellRenderer {
        
        public Component getTableCellRendererComponent(JTable table, 
                                                       Object value, 
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int row, int col) {
            JTextArea textArea = new JTextArea();
            // modify properties of text area here
            textArea.setLineWrap(true);
            textArea.setWrapStyleWord(true);

            if (value != null)
                textArea.setText(value.toString());

            // autosize the row height to fit text exactly
            textArea.setSize(table.getColumnModel().getColumn(col).getWidth(), Integer.MAX_VALUE);
            int desiredHeight = (int) textArea.getPreferredSize().getHeight();
            if (desiredHeight > table.getRowHeight(row))
                table.setRowHeight(row,desiredHeight);
            
            return textArea;
        }
    }

}
